/************************************************************************
 * Realport daemon for linux.
 *
 *  Originally written for HP-UX by Gene Olson.
 *
 *  Ported to linux by James Puzzo, 8/99
 ************************************************************************/

char copyright[] =
    "Copyright (c) 1998 Digi International.  All Rights Reserved.";

char ident[] = "$Id: drpd.c,v 1.8 2001/10/23 16:37:37 jamesp Exp $";

char version[] = DIGI_VERSION;

#include <sys/types.h>
#include <sys/times.h>
#include <sys/stat.h>
#include <sys/socket.h>

#include <sys/ioctl.h>

#include <netinet/in.h>
#include <netinet/tcp.h>

#include <arpa/inet.h>
#include <netdb.h>

#include <unistd.h>

#include <syslog.h>
#include <stdlib.h>
#include <signal.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <string.h>
#include <strings.h>

#include "digirp.h"


/************************************************************************
 * Command Parameters.
 ************************************************************************/

char* nodeName;				/* Realport node device name */

char* serverName;			/* Realport server name or
					   IP address */

int serverPort;				/* Realport server IP Port */

int debug;				/* Debug level */

int nondaemon = 0;	/* Don't daemonize if "nondaemon" is <> 0 */

link_t lk;

char progName[257];			/* Program name for error messages */

typedef struct speed_struct speed_t;

struct speed_struct
{
    int		spd_minDelay;		/* Minimum Delay in milliseconds */
    int		spd_maxSpeed;		/* Maximum speed in kilobits */
    int		spd_maxDelay;		/* Maximum Delay in milliseconds */
    int		spd_minSpeed;		/* Minimum speed in kilobits */
};

speed_t speed;				/* Speed structure */


#define LOGPRI (LOG_INFO | LOG_DAEMON)


/************************************************************************
 * Other data.
 ************************************************************************/

int serverFD;				/* Socket file descriptor */

int deviceFD;				/* Network device file descriptor */

int maxFD;				/* Maximum of two above */

/************************************************************************
 * Support for the assert() macro.
 ************************************************************************/

#define assert(x) if (!(x)) *(volatile long *)-1 = 0


/************************************************************************
 * Gets a temp string for use with printfs and such.
 *
 * The routine copies a string to its internal circular buffer,
 * and returns a pointer to it.  The strings in the buffer are
 * eventually re-used, and so should not have a lifetime longer
 * than a few printfs.
 ************************************************************************/

char* tempString(char* str)
{
    int size;
    static char buf[2000];
    static int ptr;

    size = strlen(str) + 1;

    if (ptr < size)
	ptr = sizeof(buf);

    ptr -= size;

    assert(ptr >= 0);

    strcpy(buf + ptr, str);

    return buf + ptr;
}


/************************************************************************
 * Gets the error string associated with Errno.
 ************************************************************************/

char* errorString()
{
    static char buf[30];

    char* s = strerror(errno);

    if (s == 0)
    {
	sprintf(buf, "Error %d", errno);

	s = tempString(buf);
    }

    return s;
}


/************************************************************************
 *  This is just here for debugging, so you have somewhere to put
 *  a breakpoint before the program exits.
 ************************************************************************/

void realExit(int status)
{
    exit(status);
}


/************************************************************************
 * Prints a daemon exit message and terminates.
 ************************************************************************/

void daemonExit(int status)
{
    syslog(LOGPRI, "%s Abnormal Daemon Exit\n", progName);

    realExit(status);
}


/************************************************************************
 * Catch the signal.
 ************************************************************************/

void catch(int number)
{
    syslog(LOGPRI, "%s Exit on interrupt %d\n", progName, number);

    realExit(1);
}


/************************************************************************
 * Decode line speed parameter.
 ************************************************************************/

int decodeSpeed(char* string)
{
    int n;
    char extra[10];

    n = sscanf(string,
	       "%d%1[^0-9 ]%d%1[^0-9 ]%d%1[^0-9 ]%d%1[^0-9 ]%d%1s",
	       &lk.lk_fast_rate,   extra,
	       &lk.lk_fast_delay,  extra,
	       &lk.lk_slow_rate,   extra,
	       &lk.lk_slow_delay,  extra,
	       &lk.lk_header_size, extra);

    if ((n & 1) == 0)
    {
	fprintf(stderr,
		"%s Speed format is 1-5 numbers separated by "
		"non-digits\n",
		progName);
	return 1;
    }

    if (lk.lk_fast_rate < 2400 || lk.lk_fast_rate >= 10000000)
    {
	fprintf(stderr,
		"%s speed parameter 1 (FAST RATE) out of range\n",
		progName);
	return 1;
    }

    if (n < 3)
    {
	lk.lk_fast_delay = 60;
    }
    else if (lk.lk_fast_delay < 0 || lk.lk_fast_delay >= 2400)
    {
	fprintf(stderr,
		"%s speed parameter 2 (FAST DELAY) out of range\n",
		progName);
	return 1;
    }

    if (n < 5)
    {
	lk.lk_slow_rate = lk.lk_fast_rate / 4;
    }
    else if (lk.lk_slow_rate < 600 ||
	     lk.lk_slow_rate > lk.lk_fast_rate)
    {
	fprintf(stderr,
		"%s speed parameter 3 (SLOW RATE) out of range\n",
		progName);
	return 1;
    }

    if (n < 7)
    {
	lk.lk_slow_delay = lk.lk_fast_delay + 300;
    }
    else if (lk.lk_slow_delay > 10000 ||
	     lk.lk_slow_delay < lk.lk_fast_delay)
    {
	fprintf(stderr,
		"%s speed parameter 4 (SLOW DELAY) out of range\n",
		progName);
	return 1;
    }

    if (n < 9)
    {
	lk.lk_header_size = 46;
    }
    else if (lk.lk_header_size < 2 || lk.lk_header_size > 128)
    {
	fprintf(stderr,
		"%s speed parameter 5 (HEADER SIZE) out of range\n",
		progName);
	return 1;
    }

    return 0;
}


/************************************************************************
 * Decode Command line arguments.
 ************************************************************************/

void decodeCommandLine(int argc, char **argv)
{
    int c;
    int ac;
    char extra[10];

    serverPort = 771;

    while ((c = getopt(argc, argv, "?hdnp:s:V")) != -1)
    {
	switch (c)
	{
	    /*
	     * Increment debug level.
	     */

	case 'd':
	    debug++;
	    break;

	    /*
	     * Run as a standard process, not a daemon.
	     */

	case 'n':
	    nondaemon = 1;
	    break;

	    /*
	     * Get Realport server port number.
	     */

	case 'p':
	    if (sscanf(optarg, "%d%c", &serverPort, extra) != 1 ||
		serverPort < 20 ||
		serverPort >= 65536)
	    {
		fprintf(stderr, "%s Invalid server port number\n", progName);
		goto usage;
	    }
	    break;

	    /*
	     * Check validity of line speed parameter and exit.
	     */

	case 's':
	    realExit(decodeSpeed(optarg));

	    /*
	     * Display Version and exit.
	     */

	case 'V':
	    fprintf(stderr, "%s version: %s\n", progName, version);
	    realExit(0);

	default:
	    goto usage;
	}
    }
	
    /*
     * There must be either 2 or 3 positional arguments.
     */

    ac = argc - optind;

    if (ac < 2 || ac > 3)
	goto usage;

    /*
     * Get Realport network device number.
     */

    nodeName = argv[optind];

    /*
     * Get Realport node name.
     */

    serverName = argv[optind + 1];

    /*
     * Get speed information.
     */

    if (ac >= 3 && decodeSpeed(argv[optind + 2]))
	goto usage;
	
    return;

    /*
     * Print usage message.
     */

 usage:
    fprintf(stderr,
	    "usage: %s [-?hV] [-p serverPort] "
	    "nodeName nodeAddress [speed]\n", argv[0]);
    realExit(2);
}


/************************************************************************
 * Copies data from the Network Device to the TCP connection and data
 * from the TCP connection to the Network Device.
 ************************************************************************/

void copyData()
{
    int n;
    ssize_t rcount;
    ssize_t wcount;
    ssize_t sent;
    fd_set fdread;

    u_char buf[8000];

    maxFD = serverFD;
    if (maxFD < deviceFD)
	maxFD = deviceFD;

    /*
     * Loop to copy data until an error or a disconnect.
     */

    for (;;)
    {
	/*
	 * Wait for read data to appear on either the net device
	 * or the socket.
	 */

	FD_ZERO(&fdread);

        FD_SET(serverFD, &fdread);
	FD_SET(deviceFD, &fdread);

	n = select(maxFD + 1, &fdread, (fd_set*)0, (fd_set*)0, 0);
	
	if (n <= 0)
	{
	    if (n == 0)
		continue;

	    syslog(LOGPRI, "%s select error - %m\n", progName);

	    if (debug >= 1)
	    {
		fprintf(stderr,
			"Select error - %s\n",
			errorString());
	    }
	    exit(2);
	}

	/*
	 * If server data has appeared, write it to the network
	 * device.
	 */

	if (FD_ISSET(serverFD, &fdread))
	{
	    rcount = recv(serverFD, buf, sizeof(buf), 0);

	    if (debug >= 2)
		fprintf(stderr, "Read %ld bytes from Server\n", (long)rcount);

	    if (rcount <= 0)
	    {
		syslog(LOGPRI, "%s Server disconnect\n", progName);

		if (debug >= 1)
		{
		    if (rcount == 0)
		    {
			fprintf(stderr, "Server closed connection\n");
		    }
		    else
		    {
			fprintf(stderr,
				"Server disconnect - %s\n",
				errorString());
		    }
		}

		rcount = 0;
	    }

	    wcount = write(deviceFD, buf, rcount);

	    if (wcount != rcount)
	    {
		syslog(LOGPRI, "%s Network Device write error\n", progName);

		if (debug >= 1)
		{
		    if (wcount < 0)
		    {
			fprintf(stderr,
				"Error writing %ld bytes to device - %s\n",
				(long)rcount,
				errorString());
		    }
		    else
		    {
			fprintf(stderr,
				"Incomplete write (%ld of %ld) to device\n",
				(long)wcount,
				(long)rcount);
		    }
		}

		rcount = 0;
	    }

	    if (rcount == 0)
		return;
	}

	/*
	 * When Realport data appears, write it to the network device.
	 */

	if (FD_ISSET(deviceFD, &fdread))
	{
	    rcount = read(deviceFD, buf, sizeof(buf));

	    if (rcount == 0)
		continue;

	    if (debug >= 2)
		fprintf(stderr,
			"Read %ld bytes from network device\n",
			(long)rcount);

	    if (rcount < 0)
	    {
		syslog(LOGPRI,
		       "%s Network device read error - %m\n",
		       progName);

		if (debug >= 1)
		{
		    fprintf(stderr,
			    "Network device read error - %s\n",
			    errorString());
		}
		return;
	    }

	    /*
	     * Loop to write data to the server connection until
	     * all of the data has been accepted.
	     */

	    sent = 0;

	    for (;;)
	    {
		wcount = send(serverFD, buf + sent, rcount - sent, 0);

		if (wcount <= 0)
		{
		    if (wcount < 0)
		    {
			if (errno == EAGAIN)
			{
			    if (debug >= 1)
			    {
				fprintf(stderr, "TCP link congestion\n");
			    }
			    continue;
			}

			if (debug >= 1)
			{
			    fprintf(stderr,
				    "TCP write error - %s\n",
				    errorString());
			}
		    }
		    else
		    {
			syslog(LOGPRI,
			       "%s Server disconnect (write)\n",
			       progName);

			if (debug >= 1)
			{
			    fprintf(stderr,
				    "TCP disconnect detected on write\n");
			}

			wcount = write(deviceFD, buf, 0);

			if (wcount < 0 && debug >= 1)
			{
			    fprintf(stderr,
				    "Error writing EOF to net device - %s\n",
				    errorString());
			}
		    }

		    return;
		}

		sent += wcount;

		if (sent >= rcount)
		    break;

		if (debug >= 2)
		    fprintf(stderr,
		            "Write %ld bytes to Server\n",
		            (long)wcount);
	    }

	    /*
	     * Detect a protocol error if the driver begins a packet
	     * with a RESET message.
	     */

	    if (buf[0] == 0xff)
	    {
		syslog(LOGPRI,
		       "%s Driver shutdown connection\n",
		       progName);

		if (debug >= 1)
		{
		    fprintf(stderr, "Driver requested shutdown\n");
		}

		shutdown(serverFD, 2);
		return;
	    }
	}
    }
}



/************************************************************************
 * Connects to the remote server.
 ************************************************************************/

void mainLoop()
{
    int one = 1;
    struct hostent *hp;
    struct sockaddr_in sin;
    int message = 0;
    time_t msgtime;
    time_t now;

    /*
     * Loop to handle multiple server connections.
     */

    for (;;)
    {
	/*
	 * Enter duplicate error messages into the log file
	 * once per hour.
	 */

	if (message != 0)
	{
	    time(&now);

	    if ((u_long) (now - msgtime) >= 3600)
		message = 3;
	}
	
	/*
	 * Lookup server hostname or IP address in database.
	 */

	hp = gethostbyname(serverName);

	if (hp == 0)
	{
	    if (message != 1)
	    {
		message = 1;
		time(&msgtime);

		syslog(LOGPRI,
		       "%s Server host unknown - %m\n",
		       progName);
	    }

	    if (debug >= 1)
	    {
		fprintf(stderr,
			"Server Name (%s) not found - %s\n",
			serverName, errorString());
	    }

	    sleep(600);
	    continue;
	}

	if (debug >= 1)
	{
	    fprintf(stderr, "Resolved hostname: %s\n", serverName);
	}

	/*
	 * Allocate a socket for the connection.
	 */

	serverFD = socket(AF_INET, SOCK_STREAM, 0);

	if (serverFD < 0)
	{
	    syslog(LOGPRI,
		   "%s Could not obtain socket - %m\n",
		   progName);

	    if (debug >= 1)
	    {
		fprintf(stderr,
			"Socket() error - %s\n",
			errorString());
	    }

	    daemonExit(1);
	}

	/*
	 * Connect to the server.
	 */

	memset((char *)&sin, 0, sizeof(sin));
	memcpy((char *)&sin.sin_addr, hp->h_addr, hp->h_length);

	sin.sin_family = hp->h_addrtype;
	sin.sin_port = htons((short) serverPort);
	
	if (connect(serverFD, (struct sockaddr *)&sin, sizeof(sin)) < 0)
	{
	    if (message != 2)
	    {
		message = 2;
		time(&msgtime);

		syslog(LOGPRI,
		       "%s Cannot connect to server - %m\n",
		       progName);
	    }
		
	    if (debug >= 1)
	    {
		fprintf(stderr,
			"Cannot connect to server %s:%d - %s\n",
			serverName,
			serverPort,
			errorString());
	    }
	    
	    close(serverFD);
	    
	    sleep(60);
	    continue;
	}

	if (message != 0)
	{
	    syslog(LOGPRI,
		   "%s Connected to Server\n",
		   progName);
	}
	
	if (debug >= 1)
	{
	    fprintf(stderr, "Connected to Server\n");
	}

	/*
	 * Set the TCP_NODELAY option to eliminate the
	 * default transmit delay.
	 */

	if (setsockopt(serverFD,
		       IPPROTO_TCP,
		       TCP_NODELAY,
		       &one, sizeof(one)) != 0)
	{
	    syslog(LOGPRI,
		   "%s Cannot set TCP_NDELAY - %m\n",
		   progName);

	    if (debug >= 1)
	    {
		fprintf(stderr,
			"Cannot set TCP_NODELAY: %s\n",
			errorString());
	    }
	}

	/*
	 * Copy data back and forth to the server.
	 */

	copyData();

	message = 3;

	if (debug >= 1)
	{
	    char buf[100];
	    fprintf(stdout, "Hit enter to reconnect: ");
	    fflush(stdout);
	    fgets(buf, sizeof(buf)-1, stdin);
	}
	else
	{
	    sleep(10);
	}

	/*
	 * Shut down the server connection.
	 */

	close(serverFD);
    }
}


/************************************************************************
 * Checks command parameters, opens the network device, and then
 * calls the mailLoop() procedure to do the rest.
 ************************************************************************/

int main(int argc, char **argv)
{
    char device[100];
    
    /*
     * Decode the command line.
     */

    sprintf(progName, "%s:", argv[0]);
    
    decodeCommandLine(argc, argv);

    sprintf(progName, "drpd(%s,%s)", nodeName, serverName);

    /*
     * Open the Realport network device, which then remains open
     * for the duration of the program.
     */

    sprintf(device, "/proc/dgrp/net/%s", nodeName);

    deviceFD = open(device, O_RDWR);

    if (deviceFD < 0)
    {
	syslog(LOGPRI,
	       "%s Cannot open %s - %m\n",
	       progName, device);
	       
	if (debug >= 1)
	{
	    fprintf(stderr,
		    "Cannot open %s - %s\n",
		    device, errorString());
	}

	daemonExit(1);
    }

    /*
     * If link parameters were specified, send them down to
     * the Realport device.
     */
    if (lk.lk_header_size != 0)
    {
	if (ioctl(deviceFD, DIGI_SETLINK, &lk) != 0)
	{
	    syslog(LOGPRI,
		   "%s Cannot set link parameters\n",
		   progName);

	    if (debug >= 1)
	    {
		fprintf(stderr,
			"Cannot set link parameters\n");
	    }

	    daemonExit(1);
	}
    }

    /*
     * Daemonize, unless debugging is turned on, or we have
     * explicitly requested nor to be a daemon.
     */

    if ( (debug == 0) && (nondaemon == 0) )
    {
	switch (fork())
	{
	case 0:
	    break;

	case -1:
	    syslog(LOGPRI, "%s Cannot fork - %m\n", progName);

	    fprintf(stderr, "Cannot fork - %s\n", errorString());
	    return 1;

	default:
	    return 0;
	}

	setsid();
	nice(-10);

	/*
	 * The child no longer needs "stdin", "stdout", or "stderr",
	 * and should not block processes waiting for them to close.
	 */
	fclose(stdin);
	fclose(stdout);
	fclose(stderr);
    }

    /*
     * Arrange to catch terminate interrupts.
     */

    signal(SIGINT,  catch);
    signal(SIGTERM, catch);

    /*
     * Run until interrupted.
     */

    mainLoop();

    return 0;
}
