/*****************************************************************************
 *
 * Copyright 1999 Digi International (www.digi.com)
 *     Gene Olson  <gene@digi.com>
 *     James Puzzo <jamesp@digi.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the 
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/

/************************************************************************
 * Master include file for Linux Realport Driver.
 ************************************************************************/

#define IDENT_DRP_H "$Id: drp.h,v 1.29 2001/07/13 16:53:41 dfries Exp $"

#ifndef __DRP_H
#define __DRP_H

#include <linux/types.h>
#include <asm/semaphore.h>

#include "digirp.h"

/************************************************************************
 * Tuning parameters.
 ************************************************************************/

#define CHAN_MAX	64		/* Max # ports per server */

#define SEQ_MAX		128		/* Max # transmit sequences (2^n) */
#define SEQ_MASK	(SEQ_MAX-1)	/* Sequence buffer modulus mask */

#define TBUF_MAX	4096		/* Size of transmit buffer (2^n) */
#define RBUF_MAX	4096		/* Size of receive buffer (2^n) */

#define TBUF_MASK	(TBUF_MAX-1)	/* Transmit buffer modulus mask */
#define RBUF_MASK	(RBUF_MAX-1)	/* Receive buffer modulus mask */

#define TBUF_LOW	1000		/* Transmit low water mark */

#define UIO_BASE	1000		/* Base for write operations */
#define UIO_MIN		2000		/* Minimum size application buffer */
#define UIO_MAX		8100		/* Unix I/O buffer size */

#define MON_MAX		65536		/* Monitor buffer size (2^n) */
#define MON_MASK	(MON_MAX-1)	/* Monitor wrap mask */

#define IDLE_MAX	(20 * HZ)	/* Max TCP link idle time */

#define MAX_DESC_LEN	 35	/* Maximum length of stored PS description */

/************************************************************************
 *  Forward typedefs.
 ************************************************************************/

typedef struct nd_struct nd_t;
typedef struct ch_struct ch_t;
typedef struct un_struct un_t;

/************************************************************************
 * Minor device decoding conventions.
 ************************************************************************
 *
 * For Linux, the net and mon devices are handled via "proc", so we
 * only have to mux the "tty" devices.  Since every PortServer will
 * have an individual major number, the PortServer number does not
 * need to be encoded, and in fact, does not need to exist.
 *
 */

/*
 * Port device decoding conventions:
 *
 *	Device 00 - 3f        64 dial-in modem devices. (tty)
 *	Device 40 - 7f        64 dial-out tty devices.  (cu)
 *	Device c0 - ff        64 dial-out printer devices.
 *
 *  IS_PRINT(dev)		This is a printer device.
 *
 *  OPEN_CATEGORY(dev)		Specifies the device category.  No two
 *				devices of different categories may be open
 *				at the same time.
 *
 * The following require the category returned by OPEN_CATEGORY().
 *
 *  OPEN_WAIT_AVAIL(cat)	Waits on open until the device becomes
 *				available.  Fails if NDELAY specified.
 *
 *  OPEN_WAIT_CARRIER(cat)	Waits on open if carrier is not present.
 *				Succeeds if NDELAY is given.
 *
 *  OPEN_FORCES_CARRIER(cat)	Carrier is forced high on open.
 *
 *  ALWAYS_FORCES_CARRIER(cat)	Carrier is always forced high.
 *
 *  OPEN_SETS_CLOCAL(cat)	CLOCAL is set on open.
 *
 *  HARD_HANGUP(cat)		The device is rendered inoperative after
 *				carrier loss.  It must be closed and opened
 *				again before it can be used (ala STREAMS).
 */

#define PORT_NUM(dev)			((dev) & 0x3f)		    
						    
#define OPEN_CATEGORY(dev)		(((dev) & 0x40))
#define IS_PRINT(dev)			(((dev) & 0x80) != 0)	    
			    
#define OPEN_WAIT_AVAIL(cat)		(((cat) & 0x40) == 0x000)
#define OPEN_WAIT_CARRIER(cat)  	(((cat) & 0x40) == 0x000)
#define OPEN_FORCES_CARRIER(cat)	(((cat) & 0x40) != 0x000)
#define ALWAYS_FORCES_CARRIER(cat)	0
#define OPEN_SETS_CLOCAL(cat)		0
#define HARD_HANGUP(cat)		0

						    
/************************************************************************
 * Modem signal defines for 16450/16550 compatible FEP.
 * set in ch_mout, ch_mflow, ch_mlast etc
 ************************************************************************/

/* TODO : Re-verify that these modem signal definitions are correct */

#define DM_DTR		0x01
#define DM_RTS		0x02

#define DM_OUT1		0x04
#define DM_OUT2		0x08

#define DM_CTS		0x10
#define DM_DSR		0x20
#define DM_RI		0x40
#define DM_CD		0x80 		/* This is the DCD flag */


/************************************************************************
 * Realport Event Flags.
 ************************************************************************/

#define EV_OPU		0x0001		/* Ouput paused by client */
#define EV_OPS		0x0002		/* Output paused by XOFF */
#define EV_OPX		0x0004		/* Output paused by XXOFF */
#define EV_OPH		0x0008		/* Output paused by MFLOW */
#define EV_IPU		0x0010		/* Input paused by client */
#define EV_IPS		0x0020		/* Input paused by hi/low water */
#define EV_TXB		0x0040		/* Transmit break pending */
#define EV_TXI		0x0080		/* Transmit immediate pending */
#define EV_TXF		0x0100		/* Transmit flow control pending */
#define EV_RXB		0x0200		/* Break received */


/************************************************************************
 * Realport CFLAGS.
 ************************************************************************/

#define CF_CS5		0x0000		/* 5 bit characters */
#define CF_CS6		0x0010		/* 6 bit characters */
#define CF_CS7		0x0020		/* 7 bit characters */
#define CF_CS8		0x0030		/* 8 bit characters */
#define CF_CSIZE	0x0030		/* Character size */
#define CF_CSTOPB	0x0040		/* Two stop bits */
#define CF_CREAD	0x0080		/* Enable receiver */
#define CF_PARENB	0x0100		/* Enable parity */
#define CF_PARODD	0x0200		/* Odd parity */
#define CF_HUPCL	0x0400		/* Drop DTR on close */


/************************************************************************
 * Realport XFLAGS.
 ************************************************************************/

#define XF_XPAR		0x0001		/* Enable Mark/Space Parity */
#define XF_XMODEM	0x0002		/* Enable in-band modem signalling */
#define XF_XCASE	0x0004		/* Convert special characters */
#define XF_XEDATA	0x0008		/* Error data in stream */
#define XF_XTOSS	0x0010		/* Toss IXANY characters */
#define XF_XIXON	0x0020		/* xxon/xxoff enable */


/************************************************************************
 * Realport IFLAGS.
 ************************************************************************/

#define IF_IGNBRK	0x0001		/* Ignore input break */
#define IF_BRKINT	0x0002		/* Break interrupt */
#define IF_IGNPAR	0x0004		/* Ignore error characters */
#define IF_PARMRK	0x0008		/* Error chars marked with 0xff */
#define IF_INPCK	0x0010		/* Input parity checking enabled */
#define IF_ISTRIP	0x0020		/* Input chars masked with 0x7F */
#define IF_IXON		0x0400		/* Output software flow control */
#define IF_IXANY	0x0800		/* Restart output on any char */
#define	IF_IXOFF	0x1000		/* Input software flow control */
#define IF_DOSMODE	0x8000		/* 16450-compatible errors */


/************************************************************************
 * Realport OFLAGS.
 ************************************************************************/

#define OF_OLCUC	0x0002		/* Map lower to upper case */
#define OF_ONLCR	0x0004		/* Map NL to CR-NL */
#define OF_OCRNL	0x0008		/* Map CR to NL */
#define OF_ONOCR	0x0010		/* No CR output at column 0 */
#define OF_ONLRET	0x0020		/* Assume NL does NL/CR */
#define OF_TAB3		0x1800		/* Tabs expand to 8 spaces */
#define OF_TABDLY	0x1800		/* Tab delay */


/************************************************************************
 * Global locks.
 ************************************************************************/
/* TODO : historical locking placeholder */
/*
 *  In the HPUX version of the RealPort driver (which served as a basis
 *  for this driver) this locking code was used.  Saved if ever we need
 *  to review the locking under Linux.
 */
#if 0
extern lock_t drp_poll_lock;		/* Global lock */

extern b_sema_t drp_allocate_semaphore;	/* Semaphore to serialize net opens */
#endif

/************************************************************************
 * Unit flag definitions for un_flag.
 ************************************************************************/

/* These are the DIGI unit flags */
#define UN_EXCL		0x00010000	/* Exclusive open */
#define UN_STICKY	0x00020000	/* TTY Settings are now sticky */
#define UN_BUSY	    	0x00040000	/* Some work this channel */
#define UN_PWAIT    	0x00080000	/* Printer waiting for terminal */
#define UN_TIME	    	0x00100000	/* Waiting on time */
#define UN_EMPTY    	0x00200000	/* Waiting output queue empty */
#define UN_LOW	   	0x00400000	/* Waiting output low water */
#define UN_DIGI_MASK   	0x00FF0000	/* Waiting output low water */

/*
 * Definitions for async_struct (and serial_struct) flags field
 * 
 * these are the ASYNC flags copied from serial.h
 *
 */
#define UN_HUP_NOTIFY	0x0001 /* Notify getty on hangups and closes
                                   on the callout port */
#define UN_FOURPORT  	0x0002  /* Set OU1, OUT2 per AST Fourport settings */
#define UN_SAK       	0x0004  /* Secure Attention Key (Orange book) */
#define UN_SPLIT_TERMIOS 0x0008 /* Separate termios for dialin/callout */

#define UN_SPD_MASK	0x0030
#define UN_SPD_HI	0x0010  /* Use 56000 instead of 38400 bps */
#define UN_SPD_VHI	0x0020  /* Use 115200 instead of 38400 bps */
#define UN_SPD_CUST	0x0030  /* Use user-specified divisor */

#define UN_SKIP_TEST	0x0040 /* Skip UART test during autoconfiguration */
#define UN_AUTO_IRQ	0x0080 /* Do automatic IRQ during autoconfiguration */

#define UN_SESSION_LOCKOUT 0x0100 /* Lock out cua opens based on session */
#define UN_PGRP_LOCKOUT    0x0200 /* Lock out cua opens based on pgrp */
#define UN_CALLOUT_NOHUP   0x0400 /* Don't do hangups for cua device */

#define UN_FLAGS	0x0FFF  /* Possible legal async flags */
#define UN_USR_MASK	0x0430  /* Legal flags that non-privileged
                                 * users can set or reset */

#define UN_INITIALIZED		0x80000000 /* Serial port was initialized */
#define UN_CALLOUT_ACTIVE	0x40000000 /* Call out device is active */
#define UN_NORMAL_ACTIVE	0x20000000 /* Normal device is active */
#define UN_BOOT_AUTOCONF	0x10000000 /* Autoconfigure port on bootup */
#define UN_CLOSING		0x08000000 /* Serial port is closing */
#define UN_CTS_FLOW		0x04000000 /* Do CTS flow control */
#define UN_CHECK_CD		0x02000000 /* i.e., CLOCAL */
#define UN_SHARE_IRQ		0x01000000 /* for multifunction cards */


/************************************************************************
 * Structure for terminal or printer unit.  un_t
 *
 * Note that in some places the code assumes the "tty_t" is placed
 * first in the structure.
 ************************************************************************/

typedef void timeout_t(void*);

struct un_struct
{
    TTY        *un_tty;			/* System TTY struct */
    ch_t       *un_ch;			/* Associated channel */

    u_short     un_open_count;		/* Successful open count */
    u_short     un_blocked_open;	/* Reference count */
    int		un_flag;		/* Unit flags */
    u_short     un_tbusy;		/* Busy transmit count */

    WAITQ_HEAD  un_open_wait;		/* Do I really need to explain? */
    WAITQ_HEAD  un_close_wait;		/* See above explanation */
};


/************************************************************************
 * Channel State Numbers for ch_state.
 ************************************************************************/

/*
 * The ordering is important.
 *
 *    state <= CS_WAIT_CANCEL implies the channel is definitely closed.
 *
 *    state >= CS_WAIT_FAIL  implies the channel is definitely open.
 *
 *    state >= CS_READY implies data is allowed on the channel.
 */

#define CS_IDLE		0		/* Channel is idle */
#define CS_WAIT_OPEN	1		/* Waiting for Immediate Open Resp */
#define CS_WAIT_CANCEL	2		/* Waiting for Per/Incom Cancel Resp */
#define CS_WAIT_FAIL	3		/* Waiting for Immed Open Failure */
#define CS_SEND_QUERY	4		/* Ready to send Port Query */
#define CS_WAIT_QUERY	5		/* Waiting for Port Query Response */
#define CS_READY	6		/* Ready to accept commands and data */
#define CS_SEND_CLOSE	7		/* Ready to send Close Request */
#define CS_WAIT_CLOSE	8		/* Waiting for Close Response */


/************************************************************************
 * Device flag definitions for ch_flag.
 ************************************************************************/

/*
 *  Note that the state of the two carrier based flags is key.  When
 *  we check for carrier state transitions, we look at the current
 *  physical state of the DCD line and compare it with PHYS_CD (which
 *  was the state the last time we checked), and we also determine
 *  a new virtual state (composite of the physical state, FORCEDCD,
 *  CLOCAL, etc.) and compare it with VIRT_CD.
 *
 *  VIRTUAL transitions high will have the side effect of waking blocked
 *  opens.
 *
 *  PHYSICAL transitions low will cause hangups to occur _IF_ the virtual
 *  state is also low.  We DON'T want to hangup on a PURE virtual drop.
 */

#define CH_HANGUP   	0x0002   	/* Server port ready to close */

#define CH_VIRT_CD	0x0004		/* Carrier was virtually present */
#define CH_PHYS_CD	0x0008		/* Carrier was physically present */

#define CH_CLOCAL	0x0010		/* CLOCAL set in cflags */
#define CH_BAUD0	0x0020		/* Baud rate zero hangup */

#define CH_FAST_READ	0x0040		/* Fast reads are enabled */
#define CH_FAST_WRITE	0x0080		/* Fast writes are enabled */

#define CH_PRON	    	0x0100		/* Printer on string active */
#define CH_RX_FLUSH	0x0200		/* Flushing receive data */
#define CH_LOW		0x0400		/* Thread waiting for LOW water */
#define CH_EMPTY	0x0800		/* Thread waiting for EMPTY */
#define CH_DRAIN	0x1000		/* Close is waiting to drain */
#define CH_INPUT	0x2000		/* Thread waiting for INPUT */
#define CH_RXSTOP	0x4000		/* Stop output to ldisc */
#define CH_PARAM	0x8000		/* A parameter was updated */


/************************************************************************
 * Types of Open Requests for ch_otype.
 ************************************************************************/

#define OTYPE_IMMEDIATE	  0		/* Immediate Open */
#define OTYPE_PERSISTENT  1		/* Persistent Open */
#define OTYPE_INCOMING	  2		/* Incoming Open */

    
/************************************************************************
 * Request/Response flags.
 ************************************************************************/

#define RR_SEQUENCE	0x0001		/* Get server RLAST, TIN */
#define RR_STATUS	0x0002		/* Get server MINT, EINT */
#define RR_BUFFER	0x0004		/* Get server RSIZE, TSIZE */
#define RR_CAPABILITY	0x0008		/* Get server port capabilities */

#define RR_TX_FLUSH	0x0040		/* Flush output buffers */
#define RR_RX_FLUSH	0x0080		/* Flush input buffers */

#define RR_TX_STOP	0x0100		/* Pause output */
#define RR_RX_STOP	0x0200		/* Pause input */
#define RR_TX_START	0x0400		/* Start output */
#define RR_RX_START	0x0800		/* Start input */

#define RR_TX_BREAK	0x1000		/* Send BREAK */
#define RR_TX_ICHAR	0x2000		/* Send character immediate */


/************************************************************************
 * Channel information structure.   ch_t
 ************************************************************************/

struct ch_struct
{
    digi_t	ch_digi;		/* Digi variables */
    int		ch_edelay;		/* Digi edelay */

    un_t	ch_tun;			/* Terminal unit info */
    un_t	ch_pun;			/* Printer unit info */

    nd_t*	ch_nd;			/* Node pointer */
    u_char*	ch_tbuf;		/* Local Transmit Buffer */
    u_char*	ch_rbuf;		/* Local Receive Buffer */
    u_long	ch_cpstime;		/* Printer CPS time */
    u_long	ch_waketime;		/* Printer wake time */

    u_short	ch_state;		/* CS_* Protocol state */
    u_short	ch_flag;		/* CH_* flags */
    u_short	ch_send;		/* Bit vector of RR_* requests */
    u_short	ch_expect;		/* Bit vector of RR_* responses */
    u_short	ch_wait_carrier;	/* Thread count waiting for carrier */
    u_short	ch_wait_count[3];	/* Thread count waiting by otype */

    u_short	ch_open_count;		/* Successful open count */
    u_short	ch_category;		/* Device category */
    u_short	ch_open_error;		/* Last open error number */
    u_short	ch_break_time;		/* Pending break request time */
    u_short	ch_cpsrem;		/* Printer CPS remainder */
    u_short	ch_ocook;		/* Realport fastcook oflags */
    u_short	ch_inwait;		/* Thread count in CLIST input */

    u_short	ch_tin;			/* Local transmit buffer in ptr */
    u_short	ch_tout;		/* Local transmit buffer out ptr */
    u_short	ch_s_tin;		/* Realport TIN */
    u_short	ch_s_tpos;		/* Realport TPOS */
    u_short	ch_s_tsize;		/* Realport TSIZE */
    u_short	ch_s_treq;		/* Realport TREQ */
    u_short	ch_s_elast;		/* Realport ELAST */

    u_short	ch_rin;			/* Local receive buffer in ptr */
    u_short	ch_rout;		/* Local receive buffer out ptr */
    u_short	ch_s_rin;		/* Realport RIN */
	/* David Fries 7-13-2001, ch_s_rin should be renamed ch_s_rout because
	 * the variable we want to represent is the PortServer's ROUT, which is
	 * the sequence number for the next byte the PortServer will send us.
	 * RIN is the sequence number for the next byte the PortServer will
	 * receive from the uart.  The port server will send data as long as
	 * ROUT is less than RWIN.  What would happen is the port is opened, it
	 * receives data, it gives the value of RIN, we set the RWIN to
	 * RIN+RBUF_MAX-1, it sends us RWIN-ROUT bytes which overflows.  ROUT
	 * is set to zero when the port is opened, so we start at zero and
	 * count up as data is received.
	 */
    u_short	ch_s_rwin;		/* Realport RWIN */
    u_short	ch_s_rsize;		/* Realport RSIZE */

    u_short	ch_tmax;		/* Local TMAX */
    u_short	ch_ttime;		/* Local TTIME */
    u_short	ch_rmax;		/* Local RMAX */
    u_short	ch_rtime;		/* Local RTIME */
    u_short	ch_rlow;		/* Local RLOW */
    u_short	ch_rhigh;		/* Local RHIGH */

    u_short	ch_s_tmax;		/* Realport TMAX */
    u_short	ch_s_ttime;		/* Realport TTIME */
    u_short	ch_s_rmax;		/* Realport RMAX */
    u_short	ch_s_rtime;		/* Realport RTIME */
    u_short	ch_s_rlow;		/* Realport RLOW */
    u_short	ch_s_rhigh;		/* Realport RHIGH */

    u_short	ch_brate;		/* Local baud rate */
    u_short	ch_cflag;		/* Local tty cflags */
    u_short	ch_iflag;		/* Local tty iflags */
    u_short	ch_oflag;		/* Local tty oflags */
    u_short	ch_xflag;		/* Local tty xflags */

    u_short	ch_s_brate;		/* Realport BRATE */
    u_short	ch_s_cflag;		/* Realport CFLAG */
    u_short	ch_s_iflag;		/* Realport IFLAG */
    u_short	ch_s_oflag;		/* Realport OFLAG */
    u_short	ch_s_xflag;		/* Realport XFLAG */

    u_char	ch_otype;		/* Open request type */
    u_char	ch_pscan_savechar;	/* Last character read by parity scan */
    u_char	ch_pscan_state;		/* PScan State based on last 2 chars */
    u_char	ch_otype_waiting;	/* Type of open pending in server */
    u_char	ch_flush_seq;		/* Receive flush end sequence */
    u_char	ch_s_mlast;		/* Realport MLAST */

    u_char	ch_mout;		/* Local MOUT */
    u_char	ch_mflow;		/* Local MFLOW */
    u_char	ch_mctrl;		/* Local MCTRL */
    u_char	ch_xon;			/* Local XON */
    u_char	ch_xoff;		/* Local XOFF */
    u_char	ch_lnext;		/* Local LNEXT */
    u_char	ch_xxon;		/* Local XXON */
    u_char	ch_xxoff;		/* Local XXOFF */

    u_char	ch_s_mout;		/* Realport MOUT */
    u_char	ch_s_mflow;		/* Realport MFLOW */
    u_char	ch_s_mctrl;		/* Realport MCTRL */
    u_char	ch_s_xon;		/* Realport XON */
    u_char	ch_s_xoff;		/* Realport XOFF */
    u_char	ch_s_lnext;		/* Realport LNEXT */
    u_char	ch_s_xxon;		/* Realport XXON */
    u_char	ch_s_xxoff;		/* Realport XXOFF */

    WAITQ_HEAD ch_flag_wait;          /* Wait queue for ch_flag changes */

    int		ch_custom_speed; /* Realport custom speed */
};


/************************************************************************
 * Node State definitions.
 ************************************************************************/

#define NS_CLOSED	0		/* Network device is closed */
#define NS_IDLE		1		/* Network connection inactive */
#define NS_SEND_QUERY	2		/* Send server query */
#define NS_WAIT_QUERY	3		/* Wait for query response */
#define NS_READY	4		/* Network ready */
#define NS_SEND_ERROR	5		/* Must send error hangup */

#define ND_STATE_STR(x) \
	((x)==NS_CLOSED     ? "CLOSED"     : \
	((x)==NS_IDLE       ? "IDLE"       : \
	((x)==NS_SEND_QUERY ? "SEND_QUERY" : \
	((x)==NS_WAIT_QUERY ? "WAIT_QUERY" : \
	((x)==NS_READY      ? "READY"      : \
	((x)==NS_SEND_ERROR ? "SEND_ERROR" : "UNKNOWN" ))))))

/************************************************************************
 * Node Flag definitions.
 ************************************************************************/

#define ND_SELECT	0x0001		/* Multiple net read selects */
#define ND_DEB_WAIT	0x0002		/* Debug Device waiting */


/************************************************************************
 * Monitoring flag definitions.
 ************************************************************************/

#define MON_WAIT_DATA	0x0001		/* Waiting for buffer data */
#define MON_WAIT_SPACE	0x0002		/* Waiting for buffer space */


/************************************************************************
 * Definitions taken from Realport Dump.
 ************************************************************************/

#define RPDUMP_MAGIC	"Digi-RealPort-1.0"

#define RPDUMP_MESSAGE	0xE2		/* Descriptive message */
#define RPDUMP_RESET	0xE7		/* Connection reset */
#define RPDUMP_CLIENT	0xE8		/* Client data */
#define RPDUMP_SERVER	0xE9		/* Server data */


/************************************************************************
 * Node request/response definitions.
 ************************************************************************/

#define NR_ECHO		0x0001		/* Server echo packet */
#define NR_IDENT	0x0002		/* Server Product ID */
#define NR_CAPABILITY	0x0004		/* Server Capabilties */


/************************************************************************
 * Registration status of the node's Linux TTYDRIVER structures.
 ************************************************************************/
#define SERIAL_TTDRV_REG   0x0001     /* nd_serial_ttdriver registered  */
#define CALLOUT_TTDRV_REG  0x0002     /* nd_callout_ttdriver registered */
#define XPRINT_TTDRV_REG   0x0004     /* nd_xprint_ttdriver registered  */


/************************************************************************
 * Node structure.  There exists one of these for each associated
 * realport server.
 ************************************************************************/

struct nd_struct
{
	long          nd_major;            /* Node's major number           */
	long          nd_ID;               /* Node's ID code                */

	char          nd_serial_name[8];   /* ttyID plus a null.            */
	char          nd_callout_name[8];  /* cuID plus a null.             */
	char          nd_xprint_name[8];   /* xprID plus a null.            */

	int          nd_tty_ref_cnt;      /* Linux tty reference count     */

	nd_t         *nd_inext;            /* Next node in linked chain     */
	nd_t         *nd_iprev;            /* Prev node in linked chain     */
	nd_t         *nd_hnext;            /* Next node in hash table       */

	PDE          *nd_net_de;           /* Dir entry for /proc/dgrp/net  */
	PDE          *nd_mon_de;           /* Dir entry for /proc/dgrp/mon  */
	PDE          *nd_ports_de;         /* Dir entry for /proc/dgrp/ports*/

	SPINLOCK_DECLARE(nd_lock);         /* General node lock             */

	SEMAPHORE     nd_net_semaphore;    /* Net read/write lock           */
	SEMAPHORE     nd_mon_semaphore;    /* Monitor buffer lock           */

	SEMAPHORE     nd_ports_semaphore;  /* /proc/dgrp/ports exclusivity  */
	int           nd_ports_inuse;      /* use count protected by sema.  */

	int           nd_state;            /* NS_* network state            */
	int           nd_chan_count;       /* # active channels             */
	int           nd_flag;             /* Node flags                    */
	int           nd_send;             /* Responses to send             */
	int           nd_expect;           /* Responses we expect           */

	u_char       *nd_iobuf;            /* Network R/W Buffer            */
	WAITQ_HEAD    nd_tx_waitq;         /* Network select wait queue     */
	
	int           nd_tx_deposit;       /* Accumulated transmit deposits */
	int           nd_tx_charge;        /* Accumulated transmit charges  */
	int           nd_tx_credit;        /* Current TX credit             */
	int           nd_tx_ready;         /* Ready to transmit             */
	int           nd_tx_work;          /* TX work waiting               */
	u_long        nd_tx_time;          /* Last transmit time            */
	u_long        nd_poll_time;        /* Next scheduled poll time      */

	int           nd_delay;            /* Current TX delay              */
	int           nd_rate;             /* Current TX rate               */
	link_t        nd_link;             /* Link speed params.            */

	int           nd_seq_in;           /* TX seq in ptr                 */
	int           nd_seq_out;          /* TX seq out ptr                */
	int           nd_unack;            /* Unacknowledged byte count     */
	int           nd_remain;           /* Remaining receive bytes       */
	int           nd_tx_module;        /* Current TX module #           */
	int           nd_rx_module;        /* Current RX module #           */
	char         *nd_error;            /* Protocol error message        */

	int           nd_write_count;      /* drp_write() call count        */
	int           nd_read_count;       /* drp_read() count              */
	int           nd_send_count;       /* TCP message sent              */
	int           nd_tx_byte;          /* Transmit byte count           */
	int           nd_rx_byte;          /* Receive byte count            */

	u_long        nd_mon_lbolt;       /* Monitor start time             */
	int           nd_mon_flag;        /* Monitor flags                  */
	int           nd_mon_in;          /* Monitor in pointer             */
	int           nd_mon_out;         /* Monitor out pointer            */
	WAITQ_HEAD    nd_mon_wqueue;      /* Monitor wait queue (on flags)  */
	u_char       *nd_mon_buf;         /* Monitor buffer                 */

	WAITQ_HEAD    nd_seq_wque[SEQ_MAX];   /* TX thread wait queues      */
	u_char        nd_seq_wait[SEQ_MAX];   /* Transmit thread wait count */
	
	u_short       nd_seq_size[SEQ_MAX];   /* Transmit seq packet size   */
	u_long        nd_seq_time[SEQ_MAX];   /* Transmit seq packet time   */

	u_short       nd_hw_ver;           /* HW version returned from PS   */
	u_short       nd_sw_ver;           /* SW version returned from PS   */
	u_int         nd_hw_id;            /* HW ID returned from PS        */
	u_char        nd_ps_desc[MAX_DESC_LEN+1];  /* Description from PS   */

	u_long        nd_ttdriver_flags;   /* Registration status           */
	TTYDRIVER     nd_serial_ttdriver;  /* Linux TTYDRIVER structure      */
	TTYDRIVER     nd_callout_ttdriver; /* Linux TTYDRIVER structure      */
	TTYDRIVER     nd_xprint_ttdriver;  /* Linux TTYDRIVER structure      */

	ch_t          nd_chan[CHAN_MAX];  /* Channel array                  */
};


/************************************************************************
 * Defines to keep track of time and to manage the poller.
 ************************************************************************/
#define HALF_LONG	(1 << ((sizeof(u_long) * 8) - 1))

#define TimeGE(a, b) ((u_long)((a) - (b)) < HALF_LONG)

#define TimeLT(a, b) ((u_long)((a) - (b)) >= HALF_LONG)

extern int selwait;			/* Channel where select()s wait */

extern u_long drp_tick;			/* Interval between polls */

extern u_long drp_poll_time;		/* Time of next poll */

extern int drp_active_count;		/* Number of active threads */

#endif /* __DRP_H */
